
/* Chrysalide - Outil d'analyse de fichiers binaires
 * delayed.c - équivalent Python du fichier "plugins/kaitai/parsers/delayed.c"
 *
 * Copyright (C) 2019 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "delayed.h"


#include <assert.h>
#include <pygobject.h>


#include <i18n.h>
#include <plugins/pychrysalide/access.h>
#include <plugins/pychrysalide/helpers.h>
#include <plugins/pychrysalide/analysis/content.h>
#include <plugins/pychrysalide/arch/vmpa.h>
#include <plugins/yaml/python/node.h>


#include "../record.h"
#include "../scope.h"
#include "../parsers/instance.h"
#include "../../records/delayed-int.h"



CREATE_DYN_CONSTRUCTOR(record_delayed, G_TYPE_RECORD_DELAYED);

/* Initialise une instance sur la base du dérivé de GObject. */
static int py_record_delayed_init(PyObject *, PyObject *, PyObject *);

/* Lit la valeur d'un élément Kaitai entier représenté. */
static PyObject *py_record_delayed_get_value(PyObject *, void *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet à initialiser (théoriquement).                  *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Initialise une instance sur la base du dérivé de GObject.    *
*                                                                             *
*  Retour      : 0.                                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_record_delayed_init(PyObject *self, PyObject *args, PyObject *kwds)
{
    GKaitaiInstance *inst;                  /* Instance définie créatrice  */
    kaitai_scope_t *locals;                 /* Environnement local         */
    GBinContent *content;                   /* Contenu binaire analysé     */
    int ret;                                /* Bilan de lecture des args.  */
    GRecordDelayed *delayed;                /* Création GLib à transmettre */

#define RECORD_DELAYED_DOC                                                      \
    "The RecordDelayed class stores a link to an instance used to compute a"    \
    " given value."                                                             \
    "\n"                                                                        \
    "Instances can be created using the following constructor:\n"               \
    "\n"                                                                        \
    "    RecordDelayed(inst, locals, content)"                                  \
    "\n"                                                                        \
    "Where the *inst* arguments refers to a"                                    \
    " pychrysalide.plugins.kaitai.parsers.KaitaiInstance instance as the"       \
    " creator of the newly created object, *locals* points to a"                \
    " pychrysalide.plugins.kaitai.KaitaiScope structure used as current scope." \
    " The *content* argument is a pychrysalide.analysis.BinContent instance if" \
    " the delayed instance does not define a direct value."

    /* Récupération des paramètres */

    content = NULL;

    ret = PyArg_ParseTuple(args, "O&O&|O&",
                           convert_to_kaitai_instance, &inst,
                           convert_to_kaitai_scope, &locals,
                           convert_to_binary_content_or_none, &content);
    if (!ret) return -1;

    /* Initialisation d'un objet GLib */

    ret = forward_pygobjet_init(self);
    if (ret == -1) return -1;

    /* Eléments de base */

    delayed = G_RECORD_DELAYED(pygobject_get(self));

    if (!g_record_delayed_create(delayed, inst, locals, content))
    {
        PyErr_SetString(PyExc_ValueError, _("Unable to create record delayed."));
        return -1;
    }

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Lit la valeur d'un élément Kaitai entier représenté.         *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_record_delayed_get_value(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GRecordDelayed *delayed;                /* Version native de l'élément */
    resolved_value_t resolved;              /* Valeur sous forme générique */
    bool status;                            /* Bilan d'opération           */

#define RECORD_DELAYED_VALUE_ATTRIB PYTHON_GET_DEF_FULL             \
(                                                                   \
    value, py_record_delayed,                                       \
    "Carried value (as integer, bytes), or None in case of error."  \
)

    result = NULL;

    delayed = G_RECORD_DELAYED(pygobject_get(self));

    status = g_record_delayed_compute_and_aggregate_value(delayed, &resolved);

    if (status)
        switch (resolved.type)
        {
            case GVT_ERROR:
                assert(false);
                PyErr_Format(PyExc_RuntimeError,
                             _("Error got while parsing Kaitai definition should not have been exported!"));
                result = NULL;
                break;

            case GVT_UNSIGNED_INTEGER:
                result = PyLong_FromUnsignedLongLong(resolved.unsigned_integer);
                break;

            case GVT_SIGNED_INTEGER:
                result = PyLong_FromLongLong(resolved.signed_integer);
                break;

            case GVT_FLOAT:
                result = PyFloat_FromDouble(resolved.floating_number);
                break;

            case GVT_BOOLEAN:
                result = resolved.status ? Py_True : Py_False;
                Py_INCREF(result);
                break;

            case GVT_BYTES:
                result = PyBytes_FromStringAndSize(resolved.bytes.data, resolved.bytes.len);
                exit_szstr(&resolved.bytes);
                break;

            case GVT_ARRAY:
                result = pygobject_new(G_OBJECT(resolved.array));
                break;

            case GVT_RECORD:
                result = pygobject_new(G_OBJECT(resolved.record));
                break;

            case GVT_STREAM:
                result = pygobject_new(G_OBJECT(resolved.stream));
                break;

        }

    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_record_delayed_type(void)
{
    static PyMethodDef py_record_delayed_methods[] = {
        { NULL }
    };

    static PyGetSetDef py_record_delayed_getseters[] = {
        RECORD_DELAYED_VALUE_ATTRIB,
        { NULL }
    };

    static PyTypeObject py_record_delayed_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.plugins.kaitai.records.RecordDelayed",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = RECORD_DELAYED_DOC,

        .tp_methods     = py_record_delayed_methods,
        .tp_getset      = py_record_delayed_getseters,

        .tp_init        = py_record_delayed_init,
        .tp_new         = py_record_delayed_new,

    };

    return &py_record_delayed_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide..records.RecordDelayed.*
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool ensure_python_record_delayed_is_registered(void)
{
    PyTypeObject *type;                     /* Type Python 'RecordDelayed' */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire du module      */

    type = get_python_record_delayed_type();

    if (!PyType_HasFeature(type, Py_TPFLAGS_READY))
    {
        module = get_access_to_python_module("pychrysalide.plugins.kaitai.records");

        dict = PyModule_GetDict(module);

        if (!ensure_python_match_record_is_registered())
            return false;

        if (!register_class_for_pygobject(dict, G_TYPE_RECORD_DELAYED, type))
            return false;

    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : arg = argument quelconque à tenter de convertir.             *
*                dst = destination des valeurs récupérées en cas de succès.   *
*                                                                             *
*  Description : Tente de convertir en valeur calculée.                       *
*                                                                             *
*  Retour      : Bilan de l'opération, voire indications supplémentaires.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

int convert_to_record_delayed(PyObject *arg, void *dst)
{
    int result;                             /* Bilan à retourner           */

    result = PyObject_IsInstance(arg, (PyObject *)get_python_record_delayed_type());

    switch (result)
    {
        case -1:
            /* L'exception est déjà fixée par Python */
            result = 0;
            break;

        case 0:
            PyErr_SetString(PyExc_TypeError, "unable to convert the provided argument to record delayed");
            break;

        case 1:
            *((GRecordDelayed **)dst) = G_RECORD_DELAYED(pygobject_get(arg));
            break;

        default:
            assert(false);
            break;

    }

    return result;

}
