
/* Chrysalide - Outil d'analyse de fichiers binaires
 * matches.c - sauvegarde de correspondances de motif identifiées
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "matches.h"


#include "matches-int.h"



/* Initialise la classe des séries de correspondances de motifs. */
static void g_scan_matches_class_init(GScanMatchesClass *);

/* Initialise une instance de série de correspondances trouvées. */
static void g_scan_matches_init(GScanMatches *);

/* Supprime toutes les références externes. */
static void g_scan_matches_dispose(GScanMatches *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_matches_finalize(GScanMatches *);



/* Indique le type défini pour une série de correspondances identifiées. */
G_DEFINE_TYPE(GScanMatches, g_scan_matches, G_TYPE_OBJECT);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des séries de correspondances de motifs.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_matches_class_init(GScanMatchesClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_matches_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_matches_finalize;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance de série de correspondances trouvées.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_matches_init(GScanMatches *matches)
{
    matches->source = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_matches_dispose(GScanMatches *matches)
{
    g_clear_object(&matches->source);

    G_OBJECT_CLASS(g_scan_matches_parent_class)->dispose(G_OBJECT(matches));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_matches_finalize(GScanMatches *matches)
{
    G_OBJECT_CLASS(g_scan_matches_parent_class)->finalize(G_OBJECT(matches));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = définition de correspondance à consulter.          *
*                                                                             *
*  Description : Indique la source du motif d'origine recherché.              *
*                                                                             *
*  Retour      : Motif à l'origine des correspondances.                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GSearchPattern *g_scan_matches_get_source(const GScanMatches *matches)
{
    GSearchPattern *result;                 /* Source à retourner          */

    result = matches->source;

    g_object_ref(G_OBJECT(result));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = définition de correspondance à consulter.          *
*                                                                             *
*  Description : Dénombre les correspondances enregistrées pour un motif.     *
*                                                                             *
*  Retour      : Quantité de correspondances établies.                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

size_t g_scan_matches_count(const GScanMatches *matches)
{
    size_t result;                          /* Quantité à retourner        */
    GScanMatchesClass *class;               /* Classe à activer            */

    class = G_SCAN_MATCHES_GET_CLASS(matches);

    result = class->count(matches);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = définition de correspondance à manipuler.          *
*                fd      = canal d'écriture.                                  *
*                                                                             *
*  Description : Affiche une série de correspondances au format texte.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_scan_matches_output_to_text(const GScanMatches *matches, int fd)
{
    GScanMatchesClass *class;               /* Classe à activer            */

    class = G_SCAN_MATCHES_GET_CLASS(matches);

    class->to_text(matches, fd);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = définition de correspondance à manipuler.          *
*                                                                             *
*  Description : Convertit une série de correspondances en texte.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_scan_matches_convert_as_text(const GScanMatches *matches)
{
    /* TODO */

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches  = définition de correspondance à manipuler.         *
*                padding  = éventuel bourrage initial à placer ou NULL.       *
*                level    = profondeur actuelle.                              *
*                fd       = canal d'écriture.                                 *
*                                                                             *
*  Description : Affiche une série de correspondances au format JSON.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_scan_matches_output_to_json(const GScanMatches *matches, const sized_string_t *padding, unsigned int level, int fd)
{
    GScanMatchesClass *class;               /* Classe à activer            */

    class = G_SCAN_MATCHES_GET_CLASS(matches);

    class->to_json(matches, padding, level + 1, fd);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : matches = définition de correspondance à manipuler.          *
*                                                                             *
*  Description : Convertit une série de correspondances en JSON.              *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_scan_matches_convert_as_json(const GScanMatches *matches)
{
    /* TODO */

}
